# To import required modules:
import numpy as np
import os
import sys
from scipy.special import comb

from syssimpyplots.general import *





# Test functions in "src/functions_general.py":

def test_a_from_P():
    assert np.isclose(a_from_P(365.25, 1.), 1.) # Earth's orbit

def test_P_from_a():
    assert np.isclose(P_from_a(1., 1.), 365.25) # Earth's orbit

def test_a_from_P_from_a(seed=42):
    np.random.seed(seed)
    Mstar, a = np.random.rand(2)
    assert np.isclose(a_from_P(P_from_a(a, Mstar), Mstar), a)

def test_M_from_R_rho():
    assert np.isclose(M_from_R_rho(1.), 1., atol=1e-3) # Earth mass/radius
    assert M_from_R_rho(0.) == 0

def test_rho_from_M_R():
    assert np.isclose(rho_from_M_R(1., 1.), 5.513259, atol=1e-5) # Earth mean density

def test_tdur_circ():
    assert np.isclose(tdur_circ(365.25, 1., 1.), 12.976035, atol=1e-5) # transit duration of Earth

def test_AMD(seed=42):
    np.random.seed(seed)
    assert AMD(1e-6*np.random.rand(), np.random.rand(), 0., 0.) == 0
    assert np.isclose(AMD(1e-6, 1., 0.5, 0.), 1.339746e-7, atol=1e-5)
    assert np.isclose(AMD(1e-5, 0.5, 0.1, 0.1), 7.059299e-8, atol=1e-5)

def test_photoevap_boundary_Carrera2018():
    assert photoevap_boundary_Carrera2018(2.61, 1.)
    assert photoevap_boundary_Carrera2018(1.85, 10.) == 0
    assert photoevap_boundary_Carrera2018(1.5, 50.)

def test_incl_mult_power_law_Zhu2018():
    assert np.isclose(incl_mult_power_law_Zhu2018(5), 0.8)
    assert np.isclose(incl_mult_power_law_Zhu2018(2), 19.764235, atol=1e-5)
    assert np.isclose(incl_mult_power_law_Zhu2018(4, sigma_5=1., alpha=-2.), 1.5625, atol=1e-5)

def test_cdf_normal():
    assert np.isclose(cdf_normal(0), 0.5)
    assert np.isclose(cdf_normal(1.), 0.841345, atol=1e-5)
    assert np.isclose(cdf_normal(0.)-cdf_normal(-1.), cdf_normal(1.)-cdf_normal(0.))
    assert np.isclose(cdf_normal(0.)-cdf_normal(-3.), cdf_normal(3.)-cdf_normal(0.))

def test_cdf_empirical(seed=42):
    np.random.seed(seed)
    x_unif = np.random.rand(10_000_000) # Unif(0,1)
    x_ptls = np.random.rand(10) # randomly generated percentiles
    assert np.allclose(cdf_empirical(x_unif, x_ptls), x_ptls, atol=1e-3)

    x_norm = np.random.randn(10_000_000) # Normal(0,1)
    x_evals = 2.*np.random.rand(10) - 1. # Unif(-1,1)
    assert np.allclose(cdf_empirical(x_norm, x_evals), cdf_normal(x_evals), atol=1e-3)

def test_compute_ratios():
    x1 = np.array([0.5, 1., 2., 4.])
    x2 = np.array([88.0, 224.7, 365.2, 687.0, 4331., 10747., 30589., 59800.]) # periods (days) of Solar System planets
    x3 = np.array([-2., 1.4, 5.3, -0.33, np.pi])

    # Testing 'compute_ratios_adjacent()':
    assert np.allclose(compute_ratios_adjacent(x1), 2*np.ones(len(x1)-1))
    assert np.allclose(compute_ratios_adjacent(x1)*compute_ratios_adjacent(x1[::-1]), np.ones(len(x1)-1))
    assert np.allclose(compute_ratios_adjacent(x2), [2.553409, 1.625278, 1.881161, 6.304221, 2.481413, 2.846283, 1.954951], atol=1e-5)
    assert np.allclose(compute_ratios_adjacent(x3), [-0.7, 3.785714, -0.062264, -9.519978], atol=1e-5)
    assert len(compute_ratios_adjacent([5.])) == len(compute_ratios_adjacent([])) == 0

    # Testing 'compute_ratios_all()':
    assert np.allclose(compute_ratios_all(x1), [2, 4, 8, 2, 4, 2])
    assert len(compute_ratios_all(x2)) == comb(len(x2),2)
    assert len(compute_ratios_all([5.])) == len(compute_ratios_all([])) == 0

def test_linear_fswp_bprp(seed=42):
    np.random.seed(seed)
    fswp_med, slope = np.random.rand(), 2.*np.random.rand() - 1.
    bprp_med = 0.81
    bprp_rand = np.random.rand(1000)
    fswp_rand = linear_fswp_bprp(bprp_rand, bprp_med, fswp_med=fswp_med, slope=slope)
    assert np.isclose(linear_fswp_bprp([bprp_med], bprp_med, fswp_med=fswp_med, slope=slope), fswp_med)
    assert np.all((0. <= fswp_rand) & (fswp_rand <= 1.))

def test_bin_Nmult(seed=42):
    np.random.seed(seed)
    Nmult_rand = np.random.randint(0, 1000, 8)
    Nmult_bin5p = bin_Nmult(Nmult_rand)
    assert np.sum(Nmult_bin5p) == np.sum(Nmult_rand)
    assert np.all(Nmult_rand[:4] == Nmult_bin5p[:4])
    assert Nmult_bin5p[-1] == np.sum(Nmult_rand[4:])

def test_correlation_coefficients(seed=42):
    np.random.seed(seed)
    x, y = np.random.rand(10000), np.random.rand(10000)
    assert np.abs(Pearson_correlation_coefficient(x, y)) < 0.05
    assert np.abs(Spearman_correlation_coefficient(x, y)) < 0.05
    y = x*np.random.rand()
    assert np.isclose(Pearson_correlation_coefficient(x, y), 1.)
    assert np.isclose(Spearman_correlation_coefficient(x, y), 1.)

def test_radii_star_ratio():
    r, Rstar = np.array([0.0134, 0.0085, 0.0211]), 0.94
    mu = radii_star_ratio(r, Rstar)
    assert np.isclose(mu, 0.045745, atol=1e-5)

def test_partitioning(seed=42):
    np.random.seed(seed)
    x = np.random.rand(10)
    x_lowpart = np.array([2., 2., 2., 2.])
    x_highpart = np.array([1., 1., 1., 5.])
    assert 0 <= partitioning(x) <= 1
    assert partitioning(x_lowpart) < partitioning(x_highpart)
    assert partitioning(np.random.rand()*x_lowpart) == 0

def test_monotonicity_GF2020():
    x_pos1, x_pos2 = np.array([1.1, 1.2, 1.3, 1.4]), np.array([1., 2., 3., 4.])
    x_neg1, x_neg2 = np.array([1.2, 2., 1.5, 0.9]), np.array([2., 1.5, 1.2, 0.9])
    assert 0 < monotonicity_GF2020(x_pos1) < monotonicity_GF2020(x_pos2)
    assert monotonicity_GF2020(x_neg2) < monotonicity_GF2020(x_neg1) < 0

def test_gap_complexity_GF2020():
    x_lowgc = np.array([1., 2., 4., 8.])
    x_highgc = np.array([1., 1.2, 8., 9.5])
    assert 0 <= gap_complexity_GF2020(x_lowgc) < gap_complexity_GF2020(x_highgc) <= 1
    assert np.isclose(gap_complexity_GF2020(x_highgc), gap_complexity_GF2020(x_highgc[::-1]))



# To import required modules:
import numpy as np
import os
import sys

from syssimpyplots.general import *
from syssimpyplots.compare_kepler import *
from syssimpyplots.load_sims import *





# Test functions in "src/functions_compare_kepler.py":

def test_load_Kepler_planets_cleaned():
    pc = load_Kepler_planets_cleaned()
    assert 3. <= np.min(pc['P']) <= np.max(pc['P']) <= 300. # check range of periods (days)
    assert 0 <= np.min(pc['t_D']) <= np.max(pc['t_D']) <= 100. # check range of transit durations (hrs)
    assert 0 <= np.min(pc['depth'])/1e6 <= np.max(pc['depth'])/1e6 <= 1. # check range of transit depths
    assert 0.5 <= np.min(pc['Rp']) <= np.max(pc['Rp']) <= 10. # check range of planet radii (Earth radii)
    assert 0 < np.min(pc['Rstar'])
    assert 0 < np.min(pc['Mstar'])

def test_load_Kepler_stars_cleaned():
    sc = load_Kepler_stars_cleaned()
    assert 0 < np.min(sc['mass']) <= np.max(sc['mass']) < 2. # check range of stellar mass (Solar masses)
    assert 0 < np.min(sc['radius']) <= np.max(sc['radius']) < 4. # check range of stellar radii (Solar radii)
    assert 3000. < np.min(sc['teff']) <= np.max(sc['teff']) < 10000. # check range of T_eff (K)

def test_compute_summary_stats_from_Kepler_catalog():
    ssk_per_sys, ssk = compute_summary_stats_from_Kepler_catalog(3., 300., 0.5, 10.)
    N_sys = len(ssk_per_sys['Mtot_obs']) # total number of systems
    N_multis = np.sum(ssk['Nmult_obs'][1:]) # total number of multi-planet systems
    assert N_sys == np.sum(ssk['Nmult_obs'])
    N_pl = np.sum(ssk_per_sys['Mtot_obs']) # total number of planets
    assert N_pl == np.sum(ssk['Nmult_obs'] * np.arange(1,len(ssk['Nmult_obs'])+1))
    N_pl_pairs = np.sum(ssk['Nmult_obs'] * np.arange(len(ssk['Nmult_obs']))) # total number of adjacent planet pairs

    # Check that all the fields in 'ssk_per_sys' have the same number of systems:
    keys = ['Rstar_obs', 'Mstar_obs', 'teff_obs', 'bp_rp_obs', 'e_bp_rp_obs', 'cdpp4p5_obs', 'P_obs', 'D_obs', 'tdur_obs', 'tdur_tcirc_obs', 'radii_obs', 'Rm_obs', 'D_ratio_obs', 'xi_obs', 'radii_star_ratio']
    for key in keys:
        assert N_sys == len(ssk_per_sys[key])

    assert N_multis == len(ssk_per_sys['radii_partitioning'])
    assert N_multis == len(ssk_per_sys['radii_monotonicity'])
    assert np.sum(ssk['Nmult_obs'][2:]) == len(ssk_per_sys['gap_complexity'])

    # Check that all the fields in 'ssk_per_sys' have the same number of planets:
    assert N_pl == np.sum(ssk_per_sys['P_obs'] > 0) == np.sum(ssk_per_sys['D_obs'] > 0) == np.sum(ssk_per_sys['radii_obs'] > 0)
    assert N_pl == np.sum(ssk_per_sys['tdur_obs'] >= 0) == np.sum(ssk_per_sys['tdur_tcirc_obs'] >= 0)
    assert N_pl_pairs == np.sum(ssk_per_sys['Rm_obs'] > 0) == np.sum(ssk_per_sys['D_ratio_obs'] > 0) == np.sum(ssk_per_sys['xi_obs'] >= 0)

    # Check that all the fields in 'ssk' have the right number of planets:
    keys = ['Rstar_obs', 'Mstar_obs', 'teff_obs', 'bp_rp_obs', 'e_bp_rp_obs', 'cdpp4p5_obs', 'P_obs', 'D_obs', 'tdur_obs', 'tdur_tcirc_obs', 'radii_obs']
    for key in keys:
        assert N_pl == len(ssk[key])
    assert N_pl == len(ssk['tdur_tcirc_obs']) == len(ssk['tdur_tcirc_1_obs']) + len(ssk['tdur_tcirc_2p_obs'])
    assert N_pl == len(ssk['D_obs']) == len(ssk['D_above_obs']) + len(ssk['D_below_obs'])

    assert N_pl_pairs == len(ssk['Rm_obs'])
    assert N_pl_pairs == len(ssk['D_ratio_obs']) == len(ssk['D_ratio_above_obs']) + len(ssk['D_ratio_below_obs']) + len(ssk['D_ratio_across_obs'])
    assert N_pl_pairs == len(ssk['xi_obs']) == len(ssk['xi_res_obs']) + len(ssk['xi_nonres_obs'])
    assert len(ssk['xi_res_obs']) >= len(ssk['xi_res32_obs']) + len(ssk['xi_res21_obs'])

def test_CRPD_dist(seed=42):
    np.random.seed(seed)
    x = np.array([1205, 252, 97, 29, 7, 3])
    assert CRPD_dist(x, x) == CRPD_dist(x, x*np.random.randint(1,10)) == 0
    assert CRPD_dist(np.random.randint(1, 1000, 5), np.random.randint(1, 1000, 5)) > 0 # WARNING: this can potentially fail in edge cases!

def test_KS_dist_mult(seed=42):
    np.random.seed(seed)
    x1, x2 = np.random.randint(1, 10, 1000), np.random.randint(1, 10, 1000)
    x3 = np.random.randint(50, 100, 1000)
    assert KS_dist_mult(x1, x1)[0] == KS_dist_mult(x2, x2)[0] == KS_dist_mult(x3, x3)[0] == 0
    assert 0 <= KS_dist_mult(x1, x2)[0] <= 1
    assert np.isclose(KS_dist_mult(x1, x2)[0], KS_dist_mult(x2, x1)[0])
    assert np.isclose(KS_dist_mult(x1, x3)[0], 1)

def test_KS_dist(seed=42):
    np.random.seed(seed)
    x1, x2 = np.random.rand(1000), np.random.randn(500)
    x1b = np.random.rand(1000)
    x3 = np.random.uniform(10., 100., 300)
    assert 0 <= KS_dist(x1, x2)[0] <= 1
    assert np.isclose(KS_dist(x1, x2)[0], KS_dist(x2, x1)[0])
    assert KS_dist(x1, x1b)[0] < KS_dist(x1, x2)[0]
    assert np.isclose(KS_dist(x1, x3)[0], 1)

def test_AD_dist(seed=42):
    np.random.seed(seed)
    x1, x2 = np.random.rand(1000), np.random.randn(500)
    x1b = np.random.rand(1000)
    x3 = np.random.uniform(10., 100., 300)

    # Test 'AD_dist()':
    assert 0 <= AD_dist(x1, x2)
    assert np.isclose(AD_dist(x1, x2), AD_dist(x2, x1))
    assert AD_dist(x1, x1b) < AD_dist(x1, x2) < AD_dist(x1, x3)

    # Test that 'AD_dist2()' returns the same results as 'AD_dist()':
    assert np.isclose(AD_dist(x1, x2), AD_dist2(x1, x2))
    assert np.isclose(AD_dist(x1, x1b), AD_dist2(x1, x1b))
    assert np.isclose(AD_dist(x1, x3), AD_dist2(x1, x3))

    # Test 'AD_mod_dist()':
    assert 0 <= AD_mod_dist(x1, x2)
    assert np.isclose(AD_mod_dist(x1, x2), AD_mod_dist(x2, x1))
    assert AD_mod_dist(x1, x1b) < AD_mod_dist(x1, x2) < AD_mod_dist(x1, x3)

def test_compute_distances_sim_Kepler():
    loadfiles_directory = '/Users/hematthi/Documents/GradSchool/Research/ACI/Simulated_Data/AMD_system/Split_stars/Singles_ecc/Params11_KS/Distribute_AMD_per_mass/durations_norm_circ_singles_multis_GF2020_KS/GP_med/'
    compute_ratios = compute_ratios_adjacent
    AD_mod = True
    weights_all = load_split_stars_weights_only()
    dists_include = ['delta_f',
                     'mult_CRPD_r',
                     'periods_KS',
                     'period_ratios_KS',
                     'durations_KS',
                     'durations_norm_circ_KS',
                     'durations_norm_circ_singles_KS',
                     'durations_norm_circ_multis_KS',
                     'duration_ratios_nonmmr_KS',
                     'duration_ratios_mmr_KS',
                     'depths_KS',
                     'radius_ratios_KS',
                     'radii_partitioning_KS',
                     'radii_monotonicity_KS',
                     'gap_complexity_KS',
                     ]

    N_sim, cos_factor, P_min, P_max, radii_min, radii_max = read_targets_period_radius_bounds(loadfiles_directory + 'periods.out')

    sss_per_sys, sss = compute_summary_stats_from_cat_obs(file_name_path=loadfiles_directory, compute_ratios=compute_ratios)
    ssk_per_sys, ssk = compute_summary_stats_from_Kepler_catalog(P_min, P_max, radii_min, radii_max, compute_ratios=compute_ratios)

    dists, dists_w = compute_distances_sim_Kepler(sss_per_sys, sss, ssk_per_sys, ssk, weights_all['all'], dists_include, N_sim, cos_factor=cos_factor, AD_mod=AD_mod)

    for key in dists.keys():
        assert dists[key] >= 0 # check that all distance terms are non-negative
        if key[-2:] == 'KS':
            assert dists[key] <= 1
    for key in dists_w.keys():
        assert dists_w[key] >= 0 # check that all weighted distance terms are non-negative

    # Also implicitly tests 'compute_total_weighted_dist()':
    tot_dist_w = 0.
    for key in dists_include:
        tot_dist_w += dists_w[key]
    assert np.isclose(dists_w['tot_dist_w_include'], tot_dist_w)
